#!/usr/bin/env python3

import os
from optparse import OptionParser

from sgn.apps import Pipeline

from sgnts.sources import FakeSeriesSrc
from sgnts.sinks import DumpSeriesSink
from sgnligo.transforms import Whiten, HorizonDistance
from sgnligo.sources import FrameReader
from sgnts.transforms import Resampler
import os

def parse_command_line():
    parser = OptionParser()

    parser.add_option("--instrument", metavar = "ifo", help = "Instrument to analyze. H1, L1, or V1.")
    parser.add_option("--gps-start-time", metavar = "seconds", help="Set the start time of the segment to analyze in GPS seconds.")
    parser.add_option("--gps-end-time", metavar = "seconds", help="Set the end time of the segment to analyze in GPS seconds.")
    parser.add_option("--output-dir", metavar = "path", help = "Directory to write output data into.")
    parser.add_option("--sample-rate", metavar = "Hz", type = int, default=16384, help="Requested sampling rate of the data.")
    parser.add_option("--buffer-duration", metavar = "seconds", type = int, default = 1, help = "Length of output buffers in seconds. Default is 1 second.")
    parser.add_option("--frame-cache", metavar = "file", help="Set the path to the frame cache file to analyze.")
    parser.add_option("--channel-name", metavar = "channel", help = "Name of the data channel to analyze.")
    parser.add_option("--whitening-method", metavar = "algorithm", default = "gstlal", help = "Algorithm to use for whitening the data. Supported options are 'gwpy' or 'gstlal'. Default is gstlal.")
    parser.add_option("--reference-psd", metavar = "file", help = "load the spectrum from this LIGO light-weight XML file (optional).")
    parser.add_option("--track-psd", action = "store_true", help = "Enable dynamic PSD tracking.  Always enabled if --reference-psd is not given.")

    options, args = parser.parse_args()

    return options, args

def test_whitengraph(capsys): 

    # parse arguments
    options, args = parse_command_line()

    os.makedirs(options.output_dir, exist_ok=True)

    num_samples = options.sample_rate * options.buffer_duration

    if not (options.gps_start_time and options.gps_end_time):
        raise ValueError("Must provide both --gps-start-time and --gps-end-time.")

    # sanity check the whitening method given
    if options.whitening_method not in ("gwpy", "gstlal"):
        raise ValueError("Unknown whitening method, exiting.")

    if options.reference_psd is None:
        options.track_psd = True # FIXME not implemented

    pipeline = Pipeline()
    
    #
    #          ------   H1   -------
    #         | src1 | ---- | snk2  |
    #          ------   SR1  ------- 
    #         /
    #     H1 /
    #   ----------
    #  |  whiten  |
    #   ----------
    #          \
    #       H1  \
    #           ------ 
    #          | snk1 |
    #           ------ 
    #

    pipeline.insert(FrameReader(
               name = "FrameReader",
               source_pad_names = ("frsrc",),
               rate=options.sample_rate,
               num_samples=num_samples,
               framecache=options.frame_cache,
               channel_name = options.channel_name,
               instrument = options.instrument, 
               gps_start_time = options.gps_start_time,
               gps_end_time = options.gps_end_time,
             ),
             Resampler(
               name="Resampler",
               source_pad_names=("resamp",),
               sink_pad_names=("frsrc",),
              inrate=options.sample_rate,
               outrate=2048,
             ),
             Whiten(
               name = "Whitener",
               source_pad_names = ("hoft","spectrum"),
               sink_pad_names = ("resamp",),
               instrument = options.instrument,
               sample_rate = 2048,
               fft_length = 4,
               whitening_method = options.whitening_method,
               reference_psd = options.reference_psd,
               psd_pad_name = "Whitener:src:spectrum"
             ),
             HorizonDistance(
               name = "Horizon",
               source_pad_names = ("horizon",),
               sink_pad_names = ("spectrum",),
               m1 = 1.4,
               m2 = 1.4,
               fmin = 10.,
               fmax = 1000.,
               delta_f = 1/16.,
             ),
             DumpSeriesSink(
               name = "HoftSnk",
               sink_pad_names = ("hoft",),
               fname = os.path.join(options.output_dir, 'out.txt'),
             ),
             DumpSeriesSink(
                 name = "SpectrumSnk",
                 sink_pad_names = ("spectrum",),
                 fname = os.path.join(options.output_dir, 'psd_out.txt'),
             ),
             DumpSeriesSink(
                 name = "HorizonSnk",
                 sink_pad_names = ("horizon",),
                 fname = os.path.join(options.output_dir, "horizon.txt"),
             )
    )

    pipeline.insert(DumpSeriesSink(
           name = "RawSnk",
           sink_pad_names = ("frsrc",),
           fname = 'in.txt'
         ))
    pipeline.insert(link_map={
                              "Resampler:sink:frsrc": "FrameReader:src:frsrc",
                              "Whitener:sink:resamp": "Resampler:src:resamp",
                              "Horizon:sink:spectrum": "Whitener:src:spectrum",
                              "HoftSnk:sink:hoft": "Whitener:src:hoft",
                              "SpectrumSnk:sink:spectrum": "Whitener:src:spectrum",
                              "HorizonSnk:sink:horizon": "Horizon:src:horizon",
                              "RawSnk:sink:frsrc": "FrameReader:src:frsrc",
                              })

    pipeline.run()

if __name__ == "__main__":
    test_whitengraph(None)




import lal
from lal import LIGOTimeGPS
import lalsimulation

import math

from sgn.transforms import *
from sgnts.transforms import *
from sgnts.base import (
    SeriesBuffer,
    TSFrame,
    TSTransform,
    AdapterConfig,
)

import numpy as np

@dataclass
class HorizonDistance(TSTransform):
    """
    Compute horizon distance for an incoming PSD and a given waveform model
    """

    rate: int = 1
    fmin: int = None
    fmax: int = None
    delta_f: int = None
    m1: int = None
    m2: int = None
    spin1: tuple = (0., 0., 0.)
    spin2: tuple = (0., 0., 0.)
    eccentricity: int = 0
    inclination: int = 0
    approximant: str = "IMRPhenomD"
    range: bool = False
    snr: int = 8

    def __post_init__(self):
        super().__post_init__()

        # init waveform
        # NOTE:  the waveform models are computed up-to but not
        # including the supplied fmax parameter so we need to pass
        # (fmax + delta_f) if we want the waveform model defined
        # in the fmax bin
        hp, hc = lalsimulation.SimInspiralFD(
            self.m1 * lal.MSUN_SI, self.m2 * lal.MSUN_SI,
            self.spin1[0], self.spin1[1], self.spin1[2],
            self.spin2[0], self.spin2[1], self.spin2[2],
            1.0,    # distance (m)
            self.inclination,
            0.0,    # reference orbital phase (rad)
            0.0,    # longitude of ascending nodes (rad)
            self.eccentricity,
            0.0,    # mean anomaly of periastron
            self.delta_f,
            self.fmin,
            self.fmax + self.delta_f,
            100.,    # reference frequency (Hz)
            None,    # LAL dictionary containing accessory parameters
            lalsimulation.GetApproximantFromString(self.approximant)
        )
        assert hp.data.length > 0, "huh!?  h+ has zero length!"

        # store |h(f)|^2 for source at D = 1 m.  see (5) in
        # arXiv:1003.2481
        self.model = lal.CreateREAL8FrequencySeries(
            name = "signal spectrum",
            epoch = LIGOTimeGPS(0),
            f0 = hp.f0,
            deltaF = hp.deltaF,
            sampleUnits = hp.sampleUnits * hp.sampleUnits,
            length = hp.data.length
        )
        self.model.data.data[:] = np.abs(hp.data.data)**2.

    def compute_horizon(self, psd):
        """
        compute the horizon distance in Mpc
        """
        # frequencies at which PSD has been measured
        f = psd.f0 + np.arange(psd.data.length) * psd.deltaF

        # nearest-neighbour interpolation of waveform model
        # evaluated at PSD's frequency bins
        indexes = ((f - self.model.f0) / self.model.deltaF).round().astype("int").clip(0, self.model.data.length - 1)
        model = self.model.data.data[indexes]

        # range of indexes for integration
        kmin = (max(psd.f0, self.model.f0, self.fmin) - psd.f0) / psd.deltaF
        kmin = int(round(kmin))
        kmax = (min(psd.f0 + psd.data.length * psd.deltaF, self.model.f0 + self.model.data.length * self.model.deltaF, self.fmax) - psd.f0) / psd.deltaF
        kmax = int(round(kmax)) + 1
        assert kmin < kmax, "PSD and waveform model do not intersect"

        # SNR for source at D = 1 m <--> D in m for source w/ SNR =
        # 1.  see (3) in arXiv:1003.2481
        f = f[kmin:kmax]
        model = model[kmin:kmax]
        D = math.sqrt(4. * (model / psd.data.data[kmin:kmax]).sum() * psd.deltaF)

        # distance at desired SNR
        D /= self.snr

        # scale inspiral spectrum by distance to achieve desired SNR
        model *= 4. / D**2.

        # D in Mpc for source with specified SNR, and waveform model
        D /= (1e6 * lal.PC_SI)

        return D
         

    def transform(self, pad):
        """
        compute horizon distance
        """
        # incoming frame handling
        outbufs = []
        frame = self.preparedframes[self.sink_pads[0]]
        EOS = frame.EOS
        metadata = frame.metadata
        shape = frame.shape
        offset = frame.offset
        metadata = frame.metadata

        # get spectrum from metadata
        # FIXME: this is a hack since the PSD is a frequency series.
        psd = metadata["psd"]
        if psd is not None:
            assert isinstance(psd, lal.REAL8FrequencySeries)

            dist = self.compute_horizon(psd)
            print(f"Horizon distance: {dist} Mpc | Range = {dist / 2.25} Mpc")
        else:
            dist = None

        # send buffer with no data, put horizon history in metadata
        outbuf = SeriesBuffer(
            offset=offset, sample_rate=frame.sample_rate, data=None, shape=shape
        )
        metadata["psd_name"] = "'%s'" % pad.name
        metadata["horizon"] =  dist
        if self.range and dist is not None:
            metadata["range"] = dist/2.25

        return TSFrame(
            buffers=[outbuf],
            metadata=metadata,
            EOS=EOS,
        )



from collections import deque

from sgn.transforms import *
from sgnts.transforms import *
from sgnts.base import (
    SeriesBuffer,
    TSFrame,
    TSTransform,
    AdapterConfig,
)

from gwpy.timeseries import TimeSeries

import lal
import lal.series
from ligo.lw import utils as ligolw_utils

import numpy as np
import os
from sympy import EulerGamma
from scipy.special import loggamma
from scipy import interpolate

EULERGAMMA = float(EulerGamma.evalf())


def Y(length, i):
    """
    https://lscsoft.docs.ligo.org/lalsuite/lal/_window_8c_source.html#l00109

    Maps the length of a window and the offset within the window to the "y"
    co-ordinate of the LAL documentation.

    Input:
    length > 0,
    0 <= i < length

    Output:
    length < 2 --> return 0.0
    i == 0 --> return -1.0
    i == (length - 1) / 2 --> return 0.0
    i == length - 1 --> return +1.0

    e.g., length = 5 (odd), then i == 2 --> return 0.0
    if length = 6 (even), then i == 2.5 --> return 0.0

    (in the latter case, obviously i can't be a non-integer, but that's the
    value it would have to be for this function to return 0.0)
    """
    length -= 1
    return (2 * i - length) / length if length > 0 else 0


def interpolate_psd(
    psd: lal.REAL8FrequencySeries, deltaF: int
) -> lal.REAL8FrequencySeries:
    """Interpolates a PSD to a target frequency resolution.

    Args:
        psd:
            lal.REAL8FrequencySeries, the PSD to interpolate
        deltaF:
            int, the target frequency resolution to interpolate to

    Returns:
        lal.REAL8FrequencySeries, the interpolated PSD

    """
    # no-op?
    if deltaF == psd.deltaF:
        return psd

    # interpolate log(PSD) with cubic spline.  note that the PSD is
    # clipped at 1e-300 to prevent nan's in the interpolator (which
    # doesn't seem to like the occasional sample being -inf)
    psd_data = psd.data.data
    psd_data = np.where(psd_data, psd_data, 1e-300)
    f = psd.f0 + np.arange(len(psd_data)) * psd.deltaF
    interp = interpolate.splrep(f, np.log(psd_data), s=0)
    f = (
        psd.f0
        + np.arange(round((len(psd_data) - 1) * psd.deltaF / deltaF) + 1) * deltaF
    )
    psd_data = np.exp(interpolate.splev(f, interp, der=0))

    # return result
    psd = lal.CreateREAL8FrequencySeries(
        name=psd.name,
        epoch=psd.epoch,
        f0=psd.f0,
        deltaF=deltaF,
        sampleUnits=psd.sampleUnits,
        length=len(psd_data),
    )
    psd.data.data = psd_data

    return psd


@dataclass
class Whiten(TSTransform):
    """
    Whiten input timeseries data
    Parameters:
    -----------
    whitening-method: str
        currently supported types: (1) 'gwpy', (2) 'gstlal'
    instrument: str
        instrument to process. Used if reference-psd is given
    sample-rate: int
        sample rate of the data
    fft-length: int
        length of fft in seconds used for whitening
    nmed: int
        how many previous samples we should account for when calcualting
        the geometric mean of the psd
    navg: int
        changes to the PSD must occur over a time scale of at least
        navg*(n/2 − z)*(1/sample_rate) *check cody's paper for more info
    reference_psd: file
        path to reference psd xml
    psd_pad_name: str
        pad name of the psd output source pad
    """

    instrument: str = None
    whitening_method: str = "gwpy"
    sample_rate: int = 2048
    fft_length: int = 8
    nmed: int = 7
    navg: int = 64
    reference_psd: str = None
    psd_pad_name: str = ""

    def __post_init__(self):
        # define block overlap following arxiv:1604.04324
        self.n = int(self.fft_length * self.sample_rate)
        self.z = int(self.fft_length / 4 * self.sample_rate)
        self.hann_length = self.n - 2 * self.z
        overlap = self.hann_length // 2

        # init audio addapter
        self.adapter_config = AdapterConfig()
        self.adapter_config.overlap = (0, overlap)
        self.adapter_config.stride = self.hann_length // 2

        super().__post_init__()

        self.latest_psd = None

        # set up for gstlal method:
        if self.whitening_method == "gstlal":
            # the offset of the first output buffer
            self.first_output_offset = None

            # keep track of number of instantaneous PSDs
            # we have calculated up to navg
            self.n_samples = 0

            # set requested sampling rates
            self.delta_f = 1 / (1 / self.sample_rate) / self.n
            self.delta_t = 1 / self.sample_rate
            self.lal_normalization_constant = 2 * self.delta_f

            # store last nmed instantaneous PSDs
            self.square_data_bufs = deque(maxlen=self.nmed)
            self.prev_data = None

            # initialize window functions
            # we apply a hann window to incoming raw data
            self.hann = self.hann_window(self.n, self.z)

            # we apply a tukey window on whitened data if we have zero-padding
            if self.z:
                self.tukey = self.tukey_window(self.n, 2 * self.z / self.n)
            else:
                self.tukey = None

            # load reference PSD if provided
            if self.reference_psd:
                psd = lal.series.read_psd_xmldoc(
                    ligolw_utils.load_filename(
                        self.reference_psd,
                        verbose=True,
                        contenthandler=lal.series.PSDContentHandler,
                    )
                )
                psd = psd[self.instrument]

                # gstlal.condition
                # def psd_units_or_resolution_changed(elem, pspec, psd):
                # make sure units are set, compute scale factor
                # FIXME: what is this units?
                # units = lal.Unit(elem.get_property("psd-units"))
                # if units == lal.DimensionlessUnit:
                #    return
                # scale = float(psd.sampleUnits / units)
                scale = 1

                # get frequency resolution and number of bins
                fnyquist = self.sample_rate / 2
                n = int(round(fnyquist / self.delta_f) + 1)

                # interpolate and rescale PSD
                psd = interpolate_psd(psd, self.delta_f)
                ref_psd_data = psd.data.data[:n] * scale

                # install PSD in buffer history
                self.set_psd(ref_psd_data, self.navg)

    def tukey_window(self, length, beta):
        """
        XLALCreateTukeyREAL8Window
        https://lscsoft.docs.ligo.org/lalsuite/lal/_window_8c_source.html#l00597

        1.0 and flat in the middle, cos^2 transition at each end, zero
        at end points, 0.0 <= beta <= 1.0 sets what fraction of the
        window is transition (0 --> rectangle window, 1 --> Hann window)
        """
        if beta < 0 or beta > 1:
            raise ValueError("Invalid value for beta")

        transition_length = round(beta * length)

        n = (transition_length + 1) // 2

        out = np.ones(length)
        for i in range((transition_length + 1) // 2):
            o = np.cos(np.pi / 2 * Y(transition_length, i)) ** 2
            out[i] = o
            out[length - 1 - i] = o

        return out

    def hann_window(self, N, Z):
        """
        Define hann window
        Parameters:
        -----------
        N: int
            Number of samples in one window block
        Z: int
            Number of samples to zero pad
        """
        # array of indices
        k = np.arange(0, N, 1)

        hann = np.zeros(N)
        hann[Z : N - Z] = (np.sin(np.pi * (k[Z : N - Z] - Z) / (N - 2 * Z))) ** 2

        # FIXME gstlal had a method of adding from the two ends of the window
        # so that small numbers weren't added to big ones
        self.hann_norm = np.sqrt(N / np.sum(hann**2))

        return hann

    def median_bias(self, nn):
        """
        XLALMedianBias
        https://lscsoft.docs.ligo.org/lalsuite/lal/_average_spectrum_8c_source.html#l00378
        """
        ans = 1
        n = (nn - 1) // 2
        for i in range(1, n + 1):
            ans -= 1.0 / (2 * i)
            ans += 1.0 / (2 * i + 1)

        return ans

    def log_median_bias_geometric(self, nn):
        """
        XLALLogMedianBiasGeometric
        https://lscsoft.docs.ligo.org/lalsuite/lal/_average_spectrum_8c_source.html#l01423
        """
        return np.log(self.median_bias(nn)) - nn * (loggamma(1 / nn) - np.log(nn))

    def add_psd(self, fdata):
        """
        XLALPSDRegressorAdd
        https://lscsoft.docs.ligo.org/lalsuite/lal/_average_spectrum_8c_source.html#l01632
        """
        self.square_data_bufs.append(np.abs(fdata) ** 2)

        if self.n_samples == 0:
            self.geometric_mean_square = np.log(self.square_data_bufs[0])
            self.n_samples += 1
        else:
            self.n_samples += 1
            self.n_samples = min(self.n_samples, self.navg)
            median_bias = self.log_median_bias_geometric(len(self.square_data_bufs))

            # FIXME: this is how XLALPSDRegressorAdd gets the median,
            # but this is not exactly the median when the number is even.
            # numpy takes the average of the middle two, while this gets
            # the larger one
            log_bin_median = np.log(
                np.sort(self.square_data_bufs, axis=0)[len(self.square_data_bufs) // 2]
            )
            self.geometric_mean_square = (
                self.geometric_mean_square * (self.n_samples - 1)
                + log_bin_median
                - median_bias
            ) / self.n_samples

    def get_psd(self, fdata):
        """
        XLALPSDRegressorGetPSD
        https://lscsoft.docs.ligo.org/lalsuite/lal/_average_spectrum_8c_source.html#l01773
        """
        # running average mode (track-psd)
        if self.n_samples == 0:
            out = self.lal_normalization_constant * (np.abs(fdata) ** 2)

            # set DC and Nyquist terms to zero
            # FIXME: gstlal had a condition if self.f0 == 0
            out[0] = 0
            out[self.n // 2] = 0
            return out
        else:
            return (
                np.exp(self.geometric_mean_square + EULERGAMMA)
                * self.lal_normalization_constant
            )

    def set_psd(self, ref_psd_data, weight):
        """
        XLALPSDRegressorSetPSD
        https://lscsoft.docs.ligo.org/lalsuite/lal/_average_spectrum_8c_source.html#l01831
        """
        arithmetic_mean_square_data = ref_psd_data / self.lal_normalization_constant

        # populate the buffer history with the ref psd
        for i in range(self.nmed):
            self.square_data_bufs.append(arithmetic_mean_square_data)

        self.geometric_mean_square = np.log(arithmetic_mean_square_data) - EULERGAMMA
        self.n_samples = min(weight, self.navg)

    def transform(self, pad):
        """
        Whiten incoming data in segments of fft-length seconds overlapped by fft-length * 3/4
        If the data segment has N samples, we apply a zero-padded Hann window on the data with
        zero-padding of Z = N/4 samples. The Hann window length is then N - 2 * Z. The output
        stride is hann_length / 2.

        Example:
        --------
        fft_length = 4 sec
        sample_rate = 4
        N = 16
        Z = 4
        hann_length = 8
        output_stride = 4

        -- : input data
        .. : zero-padding
        ** : hann window
        [] : output buffer
        {} : output that will be added to the next iteration


                     *
                    * *
                   *   *
                  *     *
                 *
        1)   ....--------....
            -1s  0s      2s  3s
                 {add to next}


                         *
                        * *
                       *   *
                      *     *
                     *
        2)       ....--------....
                0s   1s      3s  4s
                [out]{add to next}


                             *
                            * *
                           *   *
                          *     *
                         *
        3)           ....--------....
                    1s   2s      4s  5s
                    [out]{add to next}


        Each fft-length of data will be windowed by the zero-padded Hann window, then
        FFTed to obtain the instantaneous PSD. The instantaneous PSD will be saved to
        a queue to calculate the running geometric mean of median PSDs, see
        arxiv:1604.04324. The running geometric mean of median PSDs from the last
        iteration will be used to whiten the current windowed-fft-length of data.
        The overlap segment from the previous output will be added to current whitened
        data. Finally the first output-stride samples of the whitened data will be put
        into the output buffer.

        Note that we will only start to produce output when the output offset is equal
        to or after the first input buffer, so the first iteration is a gap buffer.

        """
        # incoming frame handling
        outbufs = []
        frame = self.preparedframes[self.sink_pads[0]]
        EOS = frame.EOS
        metadata = frame.metadata
        outoffsets = self.preparedoutoffsets[self.sink_pads[0]]

        if self.first_output_offset is None:
            self.first_output_offset = frame.offset

        padded_data_offset = outoffsets[0]["offset"] - Offset.fromsamples(
            self.z, self.sample_rate
        )

        # FIXME: haven't tested whether this works for gwpy method
        # FIXME: can we make this more general?
        if padded_data_offset < self.first_output_offset:
            # we are in the startup stage, don't output yet
            outoffset = outoffsets[0]["offset"]
            shape = (0,)
        else:
            outoffset = padded_data_offset
            shape = (Offset.tosamples(outoffsets[0]["noffset"], self.sample_rate),)

        # passes the spectrum in metadata if the pad is the psd_pad
        if pad.name == self.psd_pad_name:
            psd = self.latest_psd
            metadata["psd"] = psd

            return TSFrame(
                buffers=[
                    SeriesBuffer(
                        offset=outoffset,
                        sample_rate=self.sample_rate,
                        data=None,
                        shape=shape,
                    )
                ],
                EOS=EOS,
                metadata=metadata,
            )

        # if audioadapter hasn't given us a frame, then we have to wait for more
        # data before we can whiten. send a gap buffer
        if frame.is_gap:
            outbufs.append(
                SeriesBuffer(
                    offset=outoffset,
                    sample_rate=self.sample_rate,
                    data=None,
                    shape=shape,
                )
            )
        else:
            # retrieve samples from the deque
            assert len(frame.buffers) == 1, "Multiple buffers not implemented yet."
            buf = frame.buffers[0]
            this_seg_data = buf.data

            if self.whitening_method == "gwpy":
                # check the type of the timeseries data.
                # transform it to a gwpy.timeseries object
                if not isinstance(this_seg_data, TimeSeries):
                    this_seg_data = TimeSeries(this_seg_data)

                # whiten it
                whitened_data = this_seg_data.whiten(
                    fftlength=self.fft_length, overlap=0, window="hann"
                )

                # transform back to a numpy array
                whitened_data = np.array(whitened_data)

            elif self.whitening_method == "gstlal":
                # apply the window function
                this_seg_data = (
                    self.hann[self.z : -self.z]
                    * this_seg_data
                    * self.delta_t
                    * self.hann_norm
                )
                this_seg_data = np.pad(this_seg_data, (self.z, self.z))

                # apply fourier transform
                freq_data = np.fft.rfft(this_seg_data)

                # get frequency bins
                freqs = np.fft.rfftfreq(this_seg_data.size, d=self.delta_t)

                # get the latest PSD
                this_psd = self.get_psd(freq_data)

                # store the latest spectrum so we can output on spectrum pad
                psd_offset = outoffsets[0]["offset"]
                psd_epoch = Offset.tosec(psd_offset)
                f0 = freqs[0]
                self.latest_psd = lal.CreateREAL8FrequencySeries(
                    "new_psd", psd_epoch, f0, self.delta_f, "s strain^2", len(this_psd)
                )
                self.latest_psd.data.data = this_psd

                # push freq data into psd history
                self.add_psd(freq_data)

                # Whitening
                # the DC and Nyquist terms are zero
                freq_data_whitened = np.zeros_like(freq_data)
                freq_data_whitened[1:-1] = freq_data[1:-1] * np.sqrt(
                    self.lal_normalization_constant / this_psd[1:-1]
                )

                # Fourier Transform back to the time domain
                # # see arxiv: 1604.04324 (13)
                # self.delta_f scaling https://lscsoft.docs.ligo.org/lalsuite/lal/_time_freq_f_f_t_8c_source.html#l00183
                whitened_data = (
                    np.fft.irfft(freq_data_whitened, self.n, norm="forward")
                    * self.delta_f
                )
                whitened_data *= self.delta_t * np.sqrt(np.sum(self.hann**2))

                if self.tukey is not None:
                    whitened_data *= self.tukey

                # accounts for overlap by summing with prev_data over the
                # stride of the adapter
                if self.prev_data is not None:
                    whitened_data[: -self.adapter_config.stride] += self.prev_data
                self.prev_data = whitened_data[self.adapter_config.stride :]

            # FIXME: haven't tested whether this works for gwpy method
            # FIXME: can we make this more general?
            if padded_data_offset < self.first_output_offset:
                # output a gap buffer during the startup period of the whitening calculation
                outdata = None
            else:
                outdata = whitened_data[: self.adapter_config.stride]

            # only output data up till the length of the adapter stride
            outbufs.append(
                SeriesBuffer(
                    offset=outoffset,
                    sample_rate=self.sample_rate,
                    data=outdata,
                    shape=shape,
                )
            )

        # return frame with the correct buffers
        return TSFrame(
            buffers=outbufs,
            EOS=EOS,
            metadata=metadata,
        )



from dataclasses import dataclass
from collections import deque, defaultdict
import yaml
from ligo.scald.io import influx

import numpy as np
from sgnts.base import TSSink

@dataclass
class InfluxSink(TSSink):
    """
    Push data to influx
    """
    instrument: str = None
    metadata_key: str = None
    scald_config: str = None
    route: str = None
    verbose: bool = False
    wait_time: float = 2

    def __post_init__(self):
        super().__post_init__()

        self.cnt = {p: 0 for p in self.sink_pads}
        self.last_reduce_time = None

        self.last_t0 = None
        # set up aggregator sink
        with open(self.scald_config, 'r') as f:
            agg_config = yaml.safe_load(f)
        self.agg_sink = influx.Aggregator(**agg_config["backends"]["default"])

        # register measurement schemas for aggregators
        self.agg_sink.load(path=self.scald_config)
        self.timedeq = deque(maxlen = 100)
        self.datadeq = {self.route: deque(maxlen = 100)}

    def pull(self, pad, bufs):
        """
        getting the buffer on the pad just modifies the name to show this final
        graph point and the prints it to prove it all works.
        """
        #super().pull(pad, bufs)
        #bufs = self.preparedframes[pad]
        self.cnt[pad] += 1
        if self.last_t0 is None:
            self.last_t0 = bufs[0].t0

        if self.metadata_key in bufs.metadata:
            # FIXME: only works when data are integers?? if float, I get `urllib3 response status: 400 | response reason: Bad Request`
            self.timedeq.append(int(bufs[0].t0/1_000_000_000))
            self.datadeq[self.route].append(int(bufs.metadata[self.metadata_key]))

        data = {self.route: {self.instrument: {'time': list(self.timedeq), 'fields': {'data': list(self.datadeq[self.route])}}}}
        if bufs[0].t0 - self.last_t0 >= int(self.wait_time*1_000_000_000):
            self.last_t0 = round(int(bufs[0].t0), -2)

            print('Writing out to influx')
            self.agg_sink.store_columns(self.route, data[self.route], aggregate="max")
            self.timedeq.clear()
            self.datadeq[self.route].clear()
        
        if bufs.EOS:
            self.mark_eos(pad)

        if self.verbose is True:
            print(self.cnt[pad], bufs)

    @property
    def EOS(self):
        """
        If buffers on any sink pads are End of Stream (EOS), then mark this whole element as EOS
        """
        return any(self.at_eos.values())





#!/usr/bin/env python3
from optparse import OptionParser

from sgn.apps import Pipeline

from sgnts.sinks import DumpSeriesSink
from sgnligo.sources import FrameReader

from sgnts.transforms import Resampler

def parse_command_line():
    parser = OptionParser()

    parser.add_option("--instrument", metavar = "ifo", help = "Instrument to analyze. H1, L1, or V1.")
    parser.add_option("--channel-name", metavar = "channel", help = "Name of the data channel to analyze.")
    parser.add_option("--gps-start-time", metavar = "seconds", help="Set the start time of the segment to analyze in GPS seconds.")
    parser.add_option("--gps-end-time", metavar = "seconds", help="Set the end time of the segment to analyze in GPS seconds.")
    parser.add_option("--frame-cache", metavar = "file", help="Set the path to the frame cache file to analyze.")
    parser.add_option("--sample-rate", metavar = "Hz", type = int, default=16384, help="Requested sampling rate of the data.")
    parser.add_option("--buffer-duration", metavar = "seconds", type = int, default = 1, help = "Length of output buffers in seconds. Default is 1 second.")

    options, args = parser.parse_args()

    return options, args

def test_framereader(capsys): 

    # parse arguments
    options, args = parse_command_line()

    if not (options.gps_start_time and options.gps_end_time):
        raise ValueError("Must provide both --gps-start-time and --gps-end-time.")

    num_samples = options.sample_rate * options.buffer_duration

    pipeline = Pipeline()

    #
    #       ---------- 
    #      | src1     |
    #       ---------- 
    #              \
    #           H1  \ SR1
    #           ------------
    #          | Resampler  |
    #           ------------
    #                 \
    #             H1   \ SR2
    #             ---------
    #            | snk1    |
    #             ---------

    pipeline.insert(FrameReader(
               name = "src1",
               source_pad_names = ("H1",),
               rate=options.sample_rate,
               num_samples=num_samples,
               framecache = options.frame_cache,
               channel_name = options.channel_name,
               instrument = options.instrument,
               gps_start_time = options.gps_start_time,
               gps_end_time = options.gps_end_time,
             ),
             Resampler(
               name="trans1",
               source_pad_names=("H1",),
               sink_pad_names=("H1",),
               inrate=options.sample_rate,
               outrate=2048,
             ),
             DumpSeriesSink(
               name = "snk1",
               sink_pad_names = ("H1",),
               fname = 'out.txt'
             )
    )

    pipeline.insert(link_map={
                              "trans1:sink:H1": "src1:src:H1",
                              "snk1:sink:H1": "trans1:src:H1"
                              })

    pipeline.run()

if __name__ == "__main__":
    test_framereader(None)






import lal
from lal import LIGOTimeGPS

import os

import time

def now():
    """
    A convenience function to return the current gps time
    """
    return LIGOTimeGPS(lal.UTCToGPS(time.gmtime()), 0)

def from_T050017(url):
    """
    Parse a URL in the style of T050017-00.
    """
    filename, _ = os.path.splitext(url)
    obs, desc, start, dur = filename.split("-")
    return obs, desc, int(start), int(dur)

def state_vector_on_off_bits(bit):
    """
    Format the given bitmask appropriately as an integer
    """
    if isinstance(bit, str):
        if not bit.startswith("0b"):
            bit = "0b" + bit
        bit = int(bit, 2)
    else:
        bit = int(bit)

    return bit




from optparse import OptionParser

from sgn.apps import Pipeline
from sgn.sinks import FakeSink

from sgnts.base import AdapterConfig, Offset
from sgnts.sources import FakeSeriesSrc
from sgnts.sinks import DumpSeriesSink, FakeSeriesSink
from sgnts.transforms import Threshold

from sgnligo.sources import DevShmSrc
from sgnligo.sinks import FakeSeriesSink, KafkaSink, InfluxSink
from sgnligo.transforms import (
    Whiten,
    Resampler,
    HorizonDistance,
)


def parse_command_line():
    parser = OptionParser(description=__doc__)

    # add our own options
    parser.add_option(
        "--sample-rate",
        metavar="Hz",
        default=2048,
        type="int",
        help="Sample rate at which to generate the PSD, default 2048 Hz",
    )
    parser.add_option(
        "--psd-fft-length",
        metavar="s",
        default=8,
        type="int",
        help="FFT length, default 8s",
    )
    parser.add_option(
        "--scald-config",
        metavar="path",
        help="sets ligo-scald options based on yaml configuration.",
    )
    parser.add_option(
        "--output-kafka-server",
        metavar="addr",
        help="Set the server address and port number for output data. Optional",
    )
    parser.add_option(
        "--analysis-tag",
        metavar="tag",
        default="test",
        help='Set the string to identify the analysis in which this job is part of. Used when --output-kafka-server is set. May not contain "." nor "-". Default is test.',
    )
    parser.add_option(
        "--reference-psd",
        metavar="filename",
        help="Load spectrum from this LIGO light-weight XML file. The noise spectrum will be measured and tracked starting from this reference. (optional).",
    )
    parser.add_option(
        "--horizon-approximant",
        type="string",
        default="IMRPhenomD",
        help="Specify a waveform approximant to use while calculating the horizon distance and range. Default is IMRPhenomD.",
    )
    parser.add_option(
        "--horizon-f-min",
        metavar="Hz",
        type="float",
        default=15.0,
        help="Set the frequency at which the waveform model is to begin for the horizon distance and range calculation. Default is 15 Hz.",
    )
    parser.add_option(
        "--horizon-f-max",
        metavar="Hz",
        type="float",
        default=900.0,
        help="Set the upper frequency cut off for the waveform model used in the horizon distance and range calculation. Default is 900 Hz.",
    )
    parser.add_option(
        "--whitening-method",
        metavar="algorithm",
        default="gstlal",
        help="Algorithm to use for whitening the data. Supported options are 'gwpy' or 'gstlal'. Default is gstlal.",
    )
    parser.add_option(
        "-v", "--verbose", action="store_true", help="Be verbose (optional)."
    )

    parser.add_option(
        "--instrument", metavar="ifo", help="Instrument to analyze. H1, L1, or V1."
    )
    parser.add_option(
        "--channel-name", metavar="channel", help="Name of the data channel to analyze."
    )
    parser.add_option(
        "--shared-memory-dir",
        metavar="directory",
        help="Set the name of the shared memory directory.",
    )
    parser.add_option(
        "--wait-time",
        metavar="seconds",
        type=int,
        default=60,
        help="Time to wait for new files in seconds before throwing an error. In online mode, new files should always arrive every second, unless there are problems. Default wait time is 60 seconds.",
    )
    parser.add_option(
        "--kafka-reduce-time",
        metavar="seconds",
        type=int,
        default=1,
        help="Time to reduce data to send to kafka.",
    )

    options, filenames = parser.parse_args()

    return options, filenames


def main():
    # parse arguments
    options, args = parse_command_line()

    source_sample_rate = 16384

    num_samples = source_sample_rate * 1

    # sanity check the whitening method given
    if options.whitening_method not in ("gwpy", "gstlal"):
        raise ValueError("Unknown whitening method, exiting.")

    if options.reference_psd is None:
        options.track_psd = True  # FIXME not implemented

    pipeline = Pipeline()

    #
    #          -----------
    #         | DevShmSrc |
    #          -----------
    #         /
    #     H1 /
    #   ------------
    #  |  Resampler |
    #   ------------
    #       |
    #   ------------  hoft ----------
    #  |  Whiten    | --- | FakeSink |
    #   ------------       ----------
    #          |psd
    #   ------------
    #  |  Horizon   |
    #   ------------
    #          \
    #       H1  \
    #           -----------
    #          | KafkaSink |
    #           -----------
    #

    pipeline.insert(
        DevShmSrc(
            name="src1",
            source_pad_names=("frsrc",),
            rate=16384,
            num_samples=16384,
            channel_name=options.channel_name,
            instrument=options.instrument,
            shared_memory_dir=options.shared_memory_dir,
            wait_time=options.wait_time,
        ),
        Resampler(
            name="Resampler",
            source_pad_names=("resamp",),
            sink_pad_names=("frsrc",),
            inrate=source_sample_rate,
            outrate=options.sample_rate,
        ),
        Whiten(
            name="Whitener",
            source_pad_names=("hoft", "spectrum"),
            sink_pad_names=("resamp",),
            instrument=options.instrument,
            sample_rate=options.sample_rate,
            fft_length=options.psd_fft_length,
            whitening_method=options.whitening_method,
            reference_psd=options.reference_psd,
            psd_pad_name="Whitener:src:spectrum",
        ),
        HorizonDistance(
            name="Horizon",
            source_pad_names=("horizon",),
            sink_pad_names=("spectrum",),
            m1=1.4,
            m2=1.4,
            fmin=options.horizon_f_min,
            fmax=options.horizon_f_max,
            range=True,
            delta_f=1 / 16.0,
        ),
        FakeSeriesSink(
            name="HoftSnk",
            sink_pad_names=("hoft",),
            verbose=True,
        ),
        KafkaSink(
            name="HorizonSnk",
            sink_pad_names=("horizon",),
            output_kafka_server=options.output_kafka_server,
            topic="gstlal."+options.analysis_tag+"."+options.instrument+"_range_history",
            route='range_history',
            metadata_key='range',
            tags=[options.instrument,],
            reduce_time=options.kafka_reduce_time,
            verbose=True
        ),
    )

    pipeline.insert(
        link_map={
            "Resampler:sink:frsrc": "src1:src:frsrc",
            "Whitener:sink:resamp": "Resampler:src:resamp",
            "Horizon:sink:spectrum": "Whitener:src:spectrum",
            "HorizonSnk:sink:horizon": "Horizon:src:horizon",
            "HoftSnk:sink:hoft": "Whitener:src:hoft",
        }
    )

    pipeline.run()


if __name__ == "__main__":
    main()



from dataclasses import dataclass

import inotify.adapters
import inotify.constants
from inotify_simple import INotify, flags
from gwpy.timeseries import TimeSeries, StateVector
import numpy as np
import os

from sgn.sources import *
from sgnts.sources import *
from .. base import *

import threading

from .utils import *

from sgnts.base.buffer import *
from sgnts.base import Offset, SeriesBuffer, TSFrame, TSSource, TSSlice, TSSlices

import queue

import time

@dataclass
class LastBuffer:
    t0: int
    is_gap: bool

@dataclass
class DevShmSrc(TSSource):
    """
    shared_memory_dir: str
        Shared memory directory name (full path).  Suggestion:  /dev/shm/kafka/L1_O3ReplayMDC
    wait_time: int
        Time to wait for next file.
    instrument: str
        instrument, should be one to one with channel names
    channel_name: tuple
        channel name of the data
    watch_suffix: str
        Filename suffix to watch for.
    """

    rate: int = 2048
    channel_name: tuple = ()
    instrument: tuple = ()
    shared_memory_dir: str = None
    wait_time: int = 60
    watch_suffix: str = ".gwf"
    state_vector_on_bits: int = None

    def __post_init__(self):
        super().__post_init__()
        self.cnt = {p: 0 for p in self.source_pads}
        self.shape = (self.num_samples,)
        self.queue = queue.Queue()

        # set assumed buffer duration based on sample rate
        # and num samples per buffer. Will fail if this does
        # not match the file duration
        self.buffer_duration = self.num_samples / self.rate

        # initialize a named tuple to track info about the previous
        # buffer sent. this will be used to make sure we dont resend
        # late data and to track discontinuities
        self.last_buffer = LastBuffer(int(now()), False)
        print(f"Start up t0: {self.last_buffer.t0}")

        # set state vector on bits
        self.bitmask = state_vector_on_off_bits(self.state_vector_on_bits)

        # Create the inotify handler
        self.observer = threading.Thread(
            target=self.monitor_dir,
            args=(self.queue, self.shared_memory_dir)
        )

        # Start the observer and set the stop attribute
        self.observer.stop = False
        self.observer.start()

    def monitor_dir(self, queue, watch_dir):
        """
        poll directory for new files with inotify
        """
        # init inotify watcher on shared memory dir
        i = INotify()
        i.add_watch(watch_dir, flags.CLOSE_WRITE | flags.MOVED_TO)

        # Get the current thread
        t = threading.currentThread()

        # Check if this thread should stop
        while not t.stop:
            # Loop over the events and check when a file has been created
            for event in i.read(timeout=1):
                # directory was removed, so the corresponding watch was
                # also removed
                if flags.IGNORED in flags.from_mask(event.mask):
                    break

                # ignore temporary files
                filename = event.name
                extension = os.path.splitext(filename)[1]
                if not (extension == self.watch_suffix):
                    continue

                # parse filename for the t0, we dont want to
                # add files to the queue if they arrive late
                _, _, t0, _ = from_T050017(filename)
                if t0 < self.last_buffer.t0:
                    pass
                else:
                    # Add the filename to the queue
                    queue.put((os.path.join(watch_dir, filename), t0))

        # Remove the watch
        i.rm_watch(watch_dir)

    def new(self, pad):
        self.cnt[pad] += 1

        # get next file from queue. if its old, try again until we
        # find a new file or reach the end of the queue
        try:
            while True:
                # Im not sure what the right timeout here is,
                # but I want to avoid a situation where get()
                # times out just before the new file arrives and
                # prematurely decides to send a gap buffer
                next_file, t0 = self.queue.get(timeout=2)
                if t0 <= self.last_buffer.t0:
                    continue
                else:
                    break

        except queue.Empty:
            if now() - self.last_buffer.t0 >= self.wait_time:
                self.observer.stop = True
                raise ValueError(f"Reached {self.wait_time} seconds with no new files in {self.shared_memory_dir}, exiting.")
            else:
                # send a gap buffer
                if self.cnt[pad] == 1:
                    # send the first gap buffer starting from the program start up time
                    t0 = self.last_buffer.t0
                else:
                    # send subsequent gaps at self.buffer_duration intervals
                    t0 = self.last_buffer.t0 + self.buffer_duration
                shape = (int(self.rate * self.buffer_duration),)
                print(f"Queue is empty, sending a gap buffer at t0: {t0}")
                outbufs = [SeriesBuffer(
                    offset=Offset.fromsec(t0 - Offset.offset_ref_t0), sample_rate=self.rate, data=None, shape=shape
                )]

                # update last buffer
                self.last_buffer.t0 = t0
        else:
            # first check the state
            statedata = StateVector.read(next_file, f"{self.instrument}:"+self.channel_name)

            state_data = np.array(statedata.data)
            state_t0 = statedata.t0.value
            state_offset0 = Offset.fromsec(state_t0 - Offset.offset_ref_t0/1_000_000_000)
            state_duration = statedata.duration.value
            state_sample_rate = statedata.sample_rate.value
            state_nsamples = int(self.rate * statedata.dt.value)

            state_times = np.arange(state_t0, state_t0 + state_duration, statedata.dt.value)
            bits = np.array(statedata)

            state_flags = []
            for b in bits:
                b = state_vector_on_off_bits(b)
                if b & self.bitmask == self.bitmask:
                    state_flags.append(True)
                else:
                    state_flags.append(False)

            if not any(state_flags):
                # return gap of buffer duration
                print(f"{self.instrument}: OFF at {t0}")

                data = None
                t0 = state_t0
                shape = (int(self.rate * self.buffer_duration),)

                outbufs = [SeriesBuffer(
                    offset=Offset.fromsec(t0 - Offset.offset_ref_t0), sample_rate=self.rate, data=data, shape=shape
                )]
            else:
                # load data from the file using gwpy
                data = TimeSeries.read(next_file, f"{self.instrument}:{self.channel_name}")
                t0 = data.t0.value
                duration = data.duration.value

                # check sample rate and duration matches what we expect
                assert int(data.sample_rate.value) == self.rate, "Data rate does not match requested sample rate."
                assert duration == self.buffer_duration, "File duration ({duration} sec) does not match assumed buffer duration ({self.buffer_duration} sec)."

                data = np.array(data)

                if all(state_flags):
                    print(f"{self.instrument}: ON at {t0}")
                    outbufs = [SeriesBuffer(
                        offset=Offset.fromsec(t0 - Offset.offset_ref_t0), sample_rate=self.rate, data=data, shape=data.shape
                    )]

                else:
                    # we need to slice the buffer to replace data with gaps
                    # for segments where state bits dont match self.bitmask
                    print(f"{self.instrument}: state transition at {t0}")

                    outbufs = []
                    state0 = state_flags[0]
                    buf_offset0 = state_offset0
                    n0 = 0
                    num_samples = state_nsamples
                    for state in state_flags[1:]:
                        if state is state0:
                            num_samples += state_nsamples
                            continue
                        else:
                            # There is a state change
                            # Create the buffer for the previous state
                            shape = (num_samples,)
                            if state0 is True:
                                outbufs.append(SeriesBuffer(
                                    offset=buf_offset0, sample_rate=self.rate, data=data[n0:n0+num_samples], shape=shape))
                            else:
                                outbufs.append(SeriesBuffer(
                                    offset=buf_offset0, sample_rate=self.rate, data=None, shape=shape))
                            # reset
                            state0 = state
                            buf_offset0 += Offset.fromsamples(num_samples, self.rate)
                            n0 += num_samples
                            num_samples = state_nsamples

                    # make the last buffer
                    if state0 is True:
                        outbufs.append(SeriesBuffer(
                            offset=buf_offset0, sample_rate=self.rate, data=data[n0:n0+num_samples], shape=(num_samples,)))
                    else:
                        outbufs.append(SeriesBuffer(
                            offset=buf_offset0, sample_rate=self.rate, data=None, shape=(num_samples,)))

            print(f"Buffer t0: {t0} | Time Now: {now()} | Time delay: {float(now()) - t0:.3e}")

            # update last buffer
            self.last_buffer.t0 = t0

        # online data is never EOS
        # FIXME but maybe there should be some kind of graceful shutdown
        EOS = False

        return TSFrame(
            buffers=outbufs,
            metadata={"cnt": self.cnt, "name": "'%s'" % pad.name},
            EOS=EOS,
        )



import sys

from .. base import *

from decimal import Decimal

from gwpy.timeseries import TimeSeries

from lal import LIGOTimeGPS
from lal.utils import CacheEntry
from ligo import segments

import numpy as np

from sgn.sources import *
from sgnts.base import Offset, SeriesBuffer, TSFrame, TSSource, TSSlice, TSSlices
from sgnts.base.buffer import *
from sgnts.sources import *


@dataclass
class FrameReader(TSSource):
    """
    channel_name: tuple
        channel names of the data
    instrument: str
        instrument, should be one to one with channel names
    framecache: path
        cache file to read data from
    gps_start_time: int
        GPS start time to analyze
    gps_end_time: int
        GPS end time to analyze
    """

    rate: int = 2048
    channel_name: str = ""
    instrument: str = ""
    framecache: str = ""
    gps_start_time: int = None
    gps_end_time: int = None

    def __post_init__(self):
        super().__post_init__()
        self.cnt = {p: 0 for p in self.source_pads}

        # init analysis segment
        self.analysis_seg = segments.segment(LIGOTimeGPS(self.gps_start_time), LIGOTimeGPS(self.gps_end_time))

        # load the cache file
        print(f"Loading {self.framecache}...")
        cache = list(map(CacheEntry, open(self.framecache)))

        # only keep files with the correct instrument
        cache = [c for c in cache if c.observatory in self.ifo_strings(self.instrument)]

        # only keep files that intersect the analysis segment
        self.cache = []
        for c in cache:
            try:
                intersection = self.analysis_seg & c.segment
            except ValueError:
                continue
            else:
                self.cache.append(c)

        # make sure it is sorted by gps time
        self.cache.sort(key=lambda x: x.segment[0])

        # init arrays for time and data
        self.offsets = np.array([])
        self.data = np.array([])

        # keep track of buffer epochs
        self.last_epoch = None

    @staticmethod
    def ifo_strings(ifo):
        """
        I dont know if the given self.instrument will be in the form of
        e.g., "H" or "H1", just make a tuple of both options for string comparison
        """
        if ifo[-1] == "1":
            return (ifo[0], ifo)
        else:
            return (ifo, ifo + "1")

    def load_gwf_data(self, frame):
        """
        load timeseries data from a gwf frame file
        """

        # get first cache entry
        segment = frame.segment

        intersection = self.analysis_seg & segment
        start = intersection[0]
        end = intersection[1]

        data = TimeSeries.read(frame.path, f"{self.instrument}:{self.channel_name}", start = start, end = end)
        assert int(data.sample_rate.value) == self.rate, "Data rate does not match requested sample rate."

        # reconstruct gps times with nanosecond precision and convert to offsets
        dt = Offset.fromsec(data.dt.value)
        gps_start = int(start.gpsSeconds * 10**9)
        gps_start_ns = int(start.gpsNanoSeconds)
        gps_end = int(end.gpsSeconds * 10**9)
        gps_end_ns = int(end.gpsNanoSeconds)

        offsets = np.arange(Offset.fromns(gps_start + gps_start_ns - Offset.offset_ref_t0), Offset.fromns(gps_end + gps_end_ns - Offset.offset_ref_t0), dt)

        data = np.array(data)

        return offsets, data

    def new(self, pad):
        """
        New buffers are created on "pad" with an instance specific count and a
        name derived from the pad name. "EOS" is set once we have procssed all data
        in the cache within the analysis segment.
        """
        self.cnt[pad] += 1
        # load next frame of data from disk when we have less than
        # one buffer length of data left
        if (self.data.size <= self.num_samples) and self.cache:
             offsets, data = self.load_gwf_data(self.cache[0])

             self.offsets = np.concatenate((self.offsets, offsets))
             self.data = np.concatenate((self.data, data))

             # now that we have loaded data from this frame,
             # remove it from the cache
             self.cache.pop(0)

        # outdata is the first self.num_samples of data in the frame
        outdata = self.data[:self.num_samples]
        outoffsets = self.offsets[:self.num_samples]

        epoch = int(outoffsets[0])
        outbuf = SeriesBuffer(
            offset=epoch, sample_rate=self.rate, data=outdata, shape=outdata.shape
        )

        # remove the used data
        self.data = self.data[self.num_samples:]
        self.offsets = self.offsets[self.num_samples:]

        # update last buffer epoch
        self.last_epoch = epoch

        # EOS condition is when we have processed all data intersecting
        # the analysis segment in every frame in the cache
        # set this condition on second to last buffer so it can propagate to
        # downstream elements
        EOS = (self.data.size <= self.num_samples) and (len(self.cache) == 0)

        return TSFrame(
            buffers=[outbuf],
            metadata={"cnt": self.cnt, "name": "'%s'" % pad.name},
            EOS=EOS
        )




from dataclasses import dataclass
from collections import deque, defaultdict

import numpy as np
from sgnts.base import TSSink

from ligo.scald.io import kafka


@dataclass
class KafkaSink(TSSink):
    """
    Push data to kafka

    Parameters:
    -----------
    output_kafka_server: str
        The kafka server to write data to
    topic: str
        The kafka topic to write data to
    metadata_key: str
        The dictionary key of the frame metadata to retrieve the data
    route:
        The route of the kafka data
    tags: str
        The tags to write the kafka data
    verbose: bool
        Be verbose
    reduce_time: float
        Will reduce data every reduce_time, in seconds
    """
    output_kafka_server: str = None
    topic: str = None
    metadata_key: str = None
    route: str = None
    tags: list = None
    verbose: bool = False
    reduce_time: float = 2

    def __post_init__(self):
        assert isinstance(self.output_kafka_server, str)
        assert isinstance(self.topic, str)
        assert isinstance(self.metadata_key, str)
        super().__post_init__()

        self.cnt = {p: 0 for p in self.sink_pads}
        self.last_reduce_time = None

        self.client = kafka.Client("kafka://{}".format(self.output_kafka_server))
        self.kafka_data = defaultdict(lambda: {'time': [], 'data': []})
        self.last_t0 = None

    def pull(self, pad, bufs):
        """
        getting the buffer on the pad just modifies the name to show this final
        graph point and the prints it to prove it all works.
        """
        self.cnt[pad] += 1
        bufst0 = bufs[0].t0/1_000_000_000
        if self.last_t0 is None:
            self.last_t0 = bufst0

        # append data to deque
        if self.metadata_key in bufs.metadata:
            self.kafka_data[self.route]['time'].append(bufst0)
            self.kafka_data[self.route]['data'].append(bufs.metadata[self.metadata_key])

        # write to kafka
        if bufst0 - self.last_t0 >= self.reduce_time:
            self.client.write(self.topic, self.kafka_data[self.route], tags=self.tags)
            self.kafka_data[self.route]['time'] = []
            self.kafka_data[self.route]['data'] = []

            self.last_t0 = bufst0
        
        if bufs.EOS:
            self.mark_eos(pad)

        if self.verbose is True:
            print(self.cnt[pad], bufs)

    @property
    def EOS(self):
        """
        If buffers on any sink pads are End of Stream (EOS), then mark this whole element as EOS
        """
        return any(self.at_eos.values())





#!/usr/bin/env python3
import os
from optparse import OptionParser

from sgn.apps import Pipeline

from sgnts.sinks import DumpSeriesSink
from sgnligo.sources import DevShmSrc

from sgnts.transforms import Resampler

def parse_command_line():
    parser = OptionParser()

    parser.add_option("--instrument", metavar = "ifo", help = "Instrument to analyze. H1, L1, or V1.")
    parser.add_option("--channel-name", metavar = "channel", help = "Name of the data channel to analyze.")
    parser.add_option("--shared-memory-dir", metavar = "directory", help = "Set the name of the shared memory directory.")
    parser.add_option("--wait-time", metavar = "seconds", type = int, default = 60, help = "Time to wait for new files in seconds before throwing an error. In online mode, new files should always arrive every second, unless there are problems. Default wait time is 60 seconds.")
    parser.add_option("--state-vector-on-bits", metavar = "bits", type = int, help = "Set the state vector on bits to process.")

    options, args = parser.parse_args()

    return options, args

def test_devshmsrc(capsys): 

    # parse arguments
    options, args = parse_command_line()

    if not os.path.exists(options.shared_memory_dir):
        raise ValueError(f"{options.shared_memory_dir} directory not found, exiting.")

    pipeline = Pipeline()

    #
    #       ---------- 
    #      | src1     |
    #       ---------- 
    #              \
    #           H1  \ SR1
    #             ---------
    #            | snk1    |
    #             ---------

    
    pipeline.insert(DevShmSrc(
               name = "src1",
               source_pad_names = ("H1",),
               rate=16384,
               num_samples=16384,
               channel_name = options.channel_name,
               instrument = options.instrument,
               shared_memory_dir = options.shared_memory_dir,
               wait_time = options.wait_time,
               state_vector_on_bits = options.state_vector_on_bits,
             ),
             DumpSeriesSink(
               name = "snk1",
               sink_pad_names = ("H1",),
               fname = 'out.txt'
             )
    )

    pipeline.insert(link_map={
                              "snk1:sink:H1": "src1:src:H1"
                              })

    pipeline.run()

if __name__ == "__main__":
    test_devshmsrc(None)



